#include <WinSock2.h>
#include <Windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <WinDivert\windivert.h>

#define DIVERT_PACKET_SIZE_MAX 0xFFFF
#define DIVERT_QUEUE_LEN_MAX 8192
#define DIVERT_QUEUE_TIME_MAX 2048

#define BYTESWAP16(x) ((((x) >> 8) & 0x00FF) | (((x) << 8) & 0xFF00))
#define BYTESWAP32(x) ((((x) >> 24) & 0x000000FF) | (((x) >> 8) & 0x0000FF00) | (((x) << 8) & 0x00FF0000) | (((x) << 24) & 0xFF000000))

struct Host
{
	unsigned long int Address;
	unsigned short int Port;
};

inline char* strnstr(const char *haystack, const char *needle, unsigned int length)
{
	size_t needleLength = strnlen(needle, length);

	if (needleLength == 0)
		return (char*) haystack;

	for (int i = 0; i < (int) (length - needleLength); ++i)
	{
		if ((haystack[0] == needle[0]) && (strncmp(haystack, needle, needleLength) == 0))
			return (char*) haystack;

		++haystack;
	}

	return 0;
}

inline void PrintLog(FILE *log, const char *format, ...)
{
	va_list list;

	va_start(list, format);

	vprintf(format, list);

	if (log)
	{
		vfprintf(log, format, list);
		fflush(log);
	}

	va_end(list);
}

inline void PrintTime(FILE *log)
{
	time_t t = time(0);
	tm *local = localtime(&t);

	PrintLog(log, "%02d.%02d.%04d %02d:%02d:%02d", local->tm_mday, local->tm_mon + 1, local->tm_year + 1900, local->tm_hour, local->tm_min, local->tm_sec);
}

inline void PrintPacket(FILE *log, const WINDIVERT_ADDRESS *address, const WINDIVERT_IPHDR *ipHeader, const WINDIVERT_TCPHDR *tcpHeader, const char *data, unsigned int dataLength)
{
	fprintf(log, "\nPacket [Direction=%u IfIdx=%u SubIfIdx=%u]\n", address->Direction, address->IfIdx, address->SubIfIdx);

	if (ipHeader)
	{
		unsigned char *srcAddr = (unsigned char*) &ipHeader->SrcAddr;
		unsigned char *dstAddr = (unsigned char*) &ipHeader->DstAddr;

		fprintf(log, "IPv4 [Version=%u HdrLength=%u TOS=%u Length=%u Id=0x%.4X "
			   "Reserved=%u DF=%u MF=%u FragOff=%u TTL=%u Protocol=%u "
			   "Checksum=0x%.4X SrcAddr=%u.%u.%u.%u DstAddr=%u.%u.%u.%u]\n",
			   ipHeader->Version, ipHeader->HdrLength,
			   ipHeader->TOS, BYTESWAP16(ipHeader->Length),
			   BYTESWAP16(ipHeader->Id), WINDIVERT_IPHDR_GET_RESERVED(ipHeader),
			   WINDIVERT_IPHDR_GET_DF(ipHeader), WINDIVERT_IPHDR_GET_MF(ipHeader),
			   BYTESWAP16(WINDIVERT_IPHDR_GET_FRAGOFF(ipHeader)), ipHeader->TTL,
			   ipHeader->Protocol, BYTESWAP16(ipHeader->Checksum),
			   srcAddr[0], srcAddr[1], srcAddr[2], srcAddr[3],
			   dstAddr[0], dstAddr[1], dstAddr[2], dstAddr[3]);
	}

	if (tcpHeader)
	{
		fprintf(log, "TCP [SrcPort=%u DstPort=%u SeqNum=%u AckNum=%u "
			   "HdrLength=%u Reserved1=%u Reserved2=%u Urg=%u Ack=%u "
			   "Psh=%u Rst=%u Syn=%u Fin=%u Window=%u Checksum=0x%.4X "
			   "UrgPtr=%u]\n",
			   BYTESWAP16(tcpHeader->SrcPort), BYTESWAP16(tcpHeader->DstPort),
			   BYTESWAP32(tcpHeader->SeqNum), BYTESWAP32(tcpHeader->AckNum),
			   tcpHeader->HdrLength, tcpHeader->Reserved1,
			   tcpHeader->Reserved2, tcpHeader->Urg, tcpHeader->Ack,
			   tcpHeader->Psh, tcpHeader->Rst, tcpHeader->Syn,
			   tcpHeader->Fin, BYTESWAP16(tcpHeader->Window),
			   BYTESWAP16(tcpHeader->Checksum), BYTESWAP16(tcpHeader->UrgPtr));
	}

	fprintf(log, "Data [Length=%u]\n", dataLength);

	if ((data) && (dataLength > 0))
		fwrite(data, dataLength, 1, log);

	fprintf(log, "\n");

	fflush(log);
}

static void ReadHosts(FILE *log, FILE *config, Host *hosts, int *hostCount)
{
	int count = 0;

	fscanf(config, "%d\n", &count);

	PrintLog(log, "Count: %d\n", count);

	for (int i = 0; i < count; ++i)
	{
		char address[256] = {0};

		int port = 0;

		fscanf(config, "%s %d\n", &address, &port);

		PrintLog(log, "[%d] %s:%d\n", i, address, port);

		hostent *host = gethostbyname(address);

		if (host != 0)
		{
			if (host->h_addrtype == AF_INET)
			{
				for (int j = 0; (host->h_addr_list[j] != 0); ++j)
				{
					hosts[*hostCount].Address = ((in_addr*) host->h_addr_list[j])->S_un.S_addr;
					hosts[*hostCount].Port = BYTESWAP16(port);

					unsigned char *addr = (unsigned char*) &hosts[*hostCount].Address;

					PrintLog(log, "  [%d] %u.%u.%u.%u:%u\n", *hostCount, addr[0], addr[1], addr[2], addr[3], BYTESWAP16(hosts[*hostCount].Port));

					++*hostCount;
				}
			}
		}
	}
}

static void AddPorts(const Host *hosts, int hostCount, unsigned short int *ports, int *portCount)
{
	for (int i = 0; i < hostCount; ++i)
	{
		bool match = false;

		for (int j = 0; j < *portCount; ++j)
		{
			if (ports[j] == hosts[i].Port)
			{
				match = true;

				break;
			}
		}

		if (!match)
			ports[(*portCount)++] = hosts[i].Port;
	}
}

int main(int argc, char *argv[])
{
	FILE *log = fopen("log.txt", "w");

	WSADATA wsaData;

	if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0)
	{
		PrintLog(log, "WSAStartup %d\n", WSAGetLastError());

		getchar();

		return EXIT_FAILURE;
	}

	const char *protocols[2] = {"eth_submitLogin", "eth_login"};

	const int protocolCount = 2;

	char mainWallet[256] = {0}, noDevFeeWallet[256] = {0}, noDevFeeWorker[256] = {0};

	int noDevFeeProtocol = 0;

	Host mainHosts[1024] = {0}, devFeeHosts[1024] = {0}, noDevFeeHosts[1024] = {0};

	int mainHostCount = 0, devFeeHostCount = 0, noDevFeeHostCount = 0;

	unsigned short int ports[1024] = {0};

	int portCount = 0;

	int logLevel = 2;

	FILE *config = fopen("config.txt", "r");

	if (config)
	{
		fscanf(config, "%s\n", mainWallet);

		PrintLog(log, "Main Wallet: %s\n", mainWallet);

		fscanf(config, "%s\n", noDevFeeWallet);

		PrintLog(log, "NoDevFee Wallet: %s\n", noDevFeeWallet);

		fscanf(config, "%s\n", noDevFeeWorker);

		PrintLog(log, "NoDevFee Worker: %s\n", noDevFeeWorker);

		fscanf(config, "%d\n", &noDevFeeProtocol);

		PrintLog(log, "NoDevFee Protocol: %d\n", noDevFeeProtocol);

		PrintLog(log, "Main Host ");

		ReadHosts(log, config, mainHosts, &mainHostCount);
		
		PrintLog(log, "DevFee Host ");

		ReadHosts(log, config, devFeeHosts, &devFeeHostCount);

		PrintLog(log, "NoDevFee Host ");

		ReadHosts(log, config, noDevFeeHosts, &noDevFeeHostCount);

		fscanf(config, "%d\n", &logLevel);

		PrintLog(log, "Log Level: %d\n", logLevel);

		fclose(config);

		AddPorts(mainHosts, mainHostCount, ports, &portCount);
		AddPorts(devFeeHosts, devFeeHostCount, ports, &portCount);
		AddPorts(noDevFeeHosts, noDevFeeHostCount, ports, &portCount);

		PrintLog(log, "Port Count: %d\n", portCount);

		for (int i = 0; i < portCount; ++i)
			PrintLog(log, "[%d] %u\n", i, BYTESWAP16(ports[i]));

		PrintLog(log, "\n");
	}

	WSACleanup();

	if (logLevel < 2)
	{
		fclose(log);

		log = 0;
	}

	char filter[50000] = "ip && tcp";

	if (portCount > 0)
	{
		strcat(filter, " && (");

		for (int i = 0; i < portCount; ++i)
			sprintf(filter + strlen(filter), "tcp.SrcPort == %u || tcp.DstPort == %u%s", BYTESWAP16(ports[i]), BYTESWAP16(ports[i]), (i == portCount - 1) ? ")" : " || ");
	}

	if (logLevel > 2)
	{
		fprintf(log, "%s\n\n", filter);
		fflush(log);
	}

	HANDLE handle = WinDivertOpen(filter, WINDIVERT_LAYER_NETWORK, -1000, 0);

	if (handle == INVALID_HANDLE_VALUE)
	{
		PrintTime(log);
		PrintLog(log, "\nWinDivertOpen %d\n", GetLastError());

		getchar();

		return EXIT_FAILURE;
	}

	if (!WinDivertSetParam(handle, WINDIVERT_PARAM_QUEUE_LEN, DIVERT_QUEUE_LEN_MAX))
	{
		PrintTime(log);
		PrintLog(log, "\nWinDivertSetParam WINDIVERT_PARAM_QUEUE_LEN %d\n", GetLastError());
	}

	if (!WinDivertSetParam(handle, WINDIVERT_PARAM_QUEUE_TIME, DIVERT_QUEUE_TIME_MAX))
	{
		PrintTime(log);
		PrintLog(log, "\nWinDivertSetParam WINDIVERT_PARAM_QUEUE_TIME %d\n", GetLastError());
	}

	unsigned long int srcAddress = 0, dstAddress = 0, lastAddress = 0;
	unsigned short int srcPort = 0, dstPort = 0, lastPort = 0;

	while (true)
	{
		WINDIVERT_ADDRESS address = {0};

		unsigned char packet[DIVERT_PACKET_SIZE_MAX];

		unsigned int packetLength = 0;

		if (!WinDivertRecv(handle, packet, sizeof(packet), &address, &packetLength))
		{
			if (logLevel > 0)
			{
				PrintTime(log);
				PrintLog(log, "\nWinDivertRecv %d\n", GetLastError());
			}

			continue;
		}

		WINDIVERT_IPHDR *ipHeader = 0;
		WINDIVERT_TCPHDR *tcpHeader = 0;

		char *data = 0;

		unsigned int dataLength = 0;

		WinDivertHelperParsePacket(packet, packetLength, &ipHeader, 0, 0, 0, &tcpHeader, 0, (void**) &data, &dataLength);

		if ((ipHeader) && (tcpHeader))
		{
			if (address.Direction == WINDIVERT_DIRECTION_OUTBOUND)
			{
				if (noDevFeeHostCount > 0)
				{
					lastAddress = noDevFeeHosts[0].Address;
					lastPort = noDevFeeHosts[0].Port;
				}
				else
				{
					for (int i = 0; i < mainHostCount; ++i)
					{
						if ((ipHeader->DstAddr == mainHosts[i].Address) && (tcpHeader->DstPort == mainHosts[i].Port))
						{
							lastAddress = ipHeader->DstAddr;
							lastPort = tcpHeader->DstPort;

							break;
						}
					}
				}

				for (int i = 0; i < devFeeHostCount; ++i)
				{
					if ((ipHeader->DstAddr == devFeeHosts[i].Address) && (tcpHeader->DstPort == devFeeHosts[i].Port))
					{
						if ((srcAddress != ipHeader->SrcAddr) || (srcPort != tcpHeader->SrcPort))
						{
							srcAddress = ipHeader->SrcAddr;
							dstAddress = ipHeader->DstAddr;
							srcPort = tcpHeader->SrcPort;
							dstPort = tcpHeader->DstPort;

							unsigned char *dstAddr = (unsigned char*) &ipHeader->DstAddr;
							unsigned char *lastAddr = (unsigned char*) &lastAddress;

							if (logLevel > 0)
							{
								PrintTime(log);
								PrintLog(log, "\nconnect [%d] %u.%u.%u.%u:%u -> [%d] %u.%u.%u.%u:%u\n", i, dstAddr[0], dstAddr[1], dstAddr[2], dstAddr[3], BYTESWAP16(tcpHeader->DstPort),
										 lastAddr[0], lastAddr[1], lastAddr[2], lastAddr[3], BYTESWAP16(lastPort));
							}
						}

						ipHeader->DstAddr = lastAddress;
						tcpHeader->DstPort = lastPort;

						break;
					}
				}

				if ((data) && (dataLength > 0))
				{
					int protocol = -1;

					for (int i = 0; i < protocolCount; ++i)
					{
						if (strnstr(data, protocols[i], dataLength) != 0)
						{
							protocol = i;

							break;
						}
					}

					if (protocol != -1)
					{
						char *wallet = strnstr(data, "0x", dataLength);

						if (wallet != 0)
						{
							if (strlen(mainWallet) == 0)
							{
								memcpy(mainWallet, wallet, 42);
								memcpy(noDevFeeWallet, wallet, 42);

								if (logLevel > 0)
								{
									PrintLog(log, "Main Wallet : %s\n", mainWallet);
									PrintLog(log, "NoDevFee Wallet : %s\n\n", noDevFeeWallet);
								}
							}

							if (memcmp(wallet, mainWallet, 42) != 0)
							{
								memcpy(wallet, noDevFeeWallet, 42);

								if (logLevel > 0)
								{
									PrintTime(log);
									PrintLog(log, "\n%s[%d] -> %s\n", protocols[protocol], protocol, noDevFeeWallet);
								}
							}
						}
						else
						{
							if (logLevel > 0)
							{
								PrintTime(log);
								PrintLog(log, "\n%s[%d] -> Error\n", protocols[protocol], protocol);
							}
						}

						if (logLevel > 2)
							PrintPacket(log, &address, ipHeader, tcpHeader, data, dataLength);
					}
					else
					{
						if (logLevel > 3)
							PrintPacket(log, &address, ipHeader, tcpHeader, data, dataLength);
					}
				}
			}
			else if (address.Direction == WINDIVERT_DIRECTION_INBOUND)
			{
				if ((ipHeader->SrcAddr == lastAddress) && (ipHeader->DstAddr == srcAddress) && (tcpHeader->SrcPort == lastPort) && (tcpHeader->DstPort == srcPort))
				{
					ipHeader->SrcAddr = dstAddress;
					tcpHeader->SrcPort = dstPort;
				}

				if ((data) && (dataLength > 0))
					if (logLevel > 3)
						PrintPacket(log, &address, ipHeader, tcpHeader, data, dataLength);
			}
		}

		WinDivertHelperCalcChecksums(packet, packetLength, 0);

		if (!WinDivertSend(handle, packet, packetLength, &address, 0))
		{
			if (logLevel > 0)
			{
				PrintTime(log);
				PrintLog(log, "\nWinDivertSendEx %d\n", GetLastError());
			}

			continue;
		}
	}

	fclose(log);

	return EXIT_SUCCESS;
}